/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "controllers/common/family/Families.h"
#include "contrib/rapidjson/document.h"
#include "contrib/rapidjson/reader.h"
#include "contrib/rapidjson/writer.h"
#include "contrib/rapidjson/prettywriter.h"
#include "Nomad3DController.h"
#include "InstrumentManager/InstrumentManager.h"
#include <core/commandzone/CommandZone.h>

using namespace std;
using namespace rapidjson;
using namespace cameo;

namespace nomad3d {

const string Nomad3DController::TYPE = "nomad3d_controller";

Nomad3DController::Nomad3DController(const string& name) :
	ExperimentController(name), controller::Init(this) {

	setFamily(family::HIDDEN);

	axisPositions.init(this, NOSAVE, "axis_positions");
}

Nomad3DController::Nomad3DController(const Nomad3DController& controller) :
	ExperimentController(controller) {
}

Nomad3DController::~Nomad3DController() {

	m_running.store(false);
	if (m_periodicUpdateThread.get() != nullptr) {
		m_periodicUpdateThread->join();
	}

	if (m_executionResponderThread.get() != nullptr) {
		m_executionResponder->cancel();
		m_executionResponderThread->join();
	}
}

void Nomad3DController::postConfiguration() {

	m_positionPublisher = application::Publisher::create("nomad3d_positions");

	m_running.store(true);
	m_periodicUpdateThread.reset(new std::thread(std::bind(&Nomad3DController::periodicUpdate, this)));

	m_executionResponder = application::Responder::create("nomad3d_execution");
	m_executionResponderThread.reset(new std::thread(std::bind(&Nomad3DController::executionLoop, this)));
}

double Nomad3DController::getUnitCoef(const std::string& unit) {

	if (unit == "deg") {
		return 1.;
	}
	if (unit == "mm") {
		return 1.;
	}
	if (unit == "cm") {
		return 10.;
	}
	if (unit == "dm") {
		return 100.;
	}
	if (unit == "m") {
		return 1000.;
	}

	cerr << "Unit " << unit << " not supported" << endl;

	return 0;
}

void Nomad3DController::init() {

	map<string, ExperimentController*> experimentControllersMap = InstrumentManager::getInstance()->getAllInstalledExperimentControllersMap();
	map<string, ExperimentController*>::iterator it;
	int i = 0;

	// Iterate the controllers.
	for (it = experimentControllersMap.begin(); it != experimentControllersMap.end(); it++) {

		// Get the AxisController controllers.
		if (dynamic_cast<axis::AxisController*>(it->second) != 0) {

			axis::AxisController* controller = dynamic_cast<axis::AxisController*>(it->second);
			controller->attach(this); //Attach it to in order to be able to receive propertyChanged events.

			AxisControllerData data;
			data.controller = controller;
			data.name = controller->getName();
			data.coef = getUnitCoef(controller->unit());

			m_axisControllers.push_back(data);

			registerUpdater(controller->position, &Nomad3DController::updateActualPosition, this, i);
			++i;
		}

		// Get the VirtualAxisController controllers.
		if (dynamic_cast<axis::VirtualAxisController*>(it->second) != 0) {

			axis::VirtualAxisController* controller = dynamic_cast<axis::VirtualAxisController*>(it->second);
			controller->attach(this); //Attach it to in order to be able to receive propertyChanged events.

			VirtualAxisControllerData data;
			data.controller = controller;
			data.name = controller->getName();
			data.coef = getUnitCoef(controller->unit());

			m_virtualAxisControllers.push_back(data);

			registerUpdater(controller->position, &Nomad3DController::updateActualPosition, this, i);
			++i;
		}
	}

	updateActualPosition(0);
}

void Nomad3DController::updateActualPosition(int32 i) {

	std::lock_guard<std::mutex> lock(m_publisherMutex);

	PrettyWriter<StringBuffer> writer(s);
	s.Clear();

	writer.StartObject();

	// Write the axis controllers.
	for (auto data : m_axisControllers) {
		writer.Key(data.name.c_str());

		// Do not apply the offset.
		float64 position = data.controller->position();

		// Get the middle position if the position is outside the limits.
		// This can be due to a unplugged encoder.
		float64 minPosition = data.controller->minPosition();
		float64 maxPosition = data.controller->maxPosition();

		if (position < minPosition || position > maxPosition) {
			position = 0.5 * (minPosition + maxPosition);
		}

		writer.Double(position * data.coef);
	}

	// Write the virtual axis controllers.
	for (auto data : m_virtualAxisControllers) {
		writer.Key(data.name.c_str());

		// Do not apply the offset.
		writer.Double(data.controller->position() * data.coef);
	}

	writer.EndObject();

	axisPositions = s.GetString();

	// Send using the publisher.
	m_positionPublisher->send(axisPositions());
}

void Nomad3DController::periodicUpdate() {

	while (m_running.load()) {

		std::this_thread::yield();
		std::this_thread::sleep_for(std::chrono::seconds(1));

		std::lock_guard<std::mutex> lock(m_publisherMutex);

		// Send using the publisher.
		m_positionPublisher->send(axisPositions());
	}
}

void Nomad3DController::executionLoop() {

	while (true) {

		unique_ptr<application::Request> request = m_executionResponder->receive();
		if (request.get() == 0) {
			break;
		}

		string requestString = request->get();

		if (requestString == "PAUSE") {

			try {
				core::CommandZone::getInstance()->pause();

			} catch (core::CommandZone::CommandZoneNotFound& e) {
				cerr << "Command zone not found" << endl;
			}
		}
		else if (requestString == "RESTART") {

			try {
				core::CommandZone::getInstance()->restart();

			} catch (core::CommandZone::CommandZoneNotFound& e) {
				cerr << "Command zone not found" << endl;
			}
		}
	}
}


}
