/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#ifndef ACQUISITION_DPPACQUISITIONCONTROLLER_H
#define ACQUISITION_DPPACQUISITIONCONTROLLER_H

#include "controllers/common/acquisition/mode/TimeAcquisitionController.h"
#include "controllers/common/acquisition/mode/DPPAcquisitionCommon.h"
#include "CAENPhaController.h"
#include "DPPLiveProcess.h"

#include <vector>
#include <boost/lexical_cast.hpp>

namespace npp {

/*!
 * \class DPPAcquisitionController
 * \brief Abstract class described the acquisition mode
 */
class DPPAcquisitionController : public acquisition::TimeAcquisitionController {

public:

	//! Type of controller
	static const std::string TYPE;

	/*!
	 * \brief Constructor
	 * \param[in] name the name of the experiment controller
	 */
	DPPAcquisitionController(const std::string& name);

	/*!
	 * \brief Destructor
	 */
	virtual ~DPPAcquisitionController();

	/*!
	 * \brief Method called before changing the property value
	 *
	 * This method is called after setting configuration during the creation of controller.
	 */
	virtual void postConfiguration();

	/*!
	 * \brief Resize the detector for z (tof) and w (slice) coordinates
	 * \throws CannotOpenFile Cannot open the file
	 * \throws SlicesFileError Slice file syntax error
	 * \throws TooManySlices Too many slices
	 * \throws SliceTooBig a slice is too big (time)
	 */
	virtual void resize() throw(CannotOpenFile, SlicesFileError, TooManySlices, SliceTooBig);

	/*!
	 * \brief Write parameters on the drivers
	 */
	virtual void writeParams();

	/*!
	 * \brief Update parameters from the drivers
	 */
	virtual void updateParams() {}

	/*!
	 * \brief Check if the acquisition using energy mode
	 * \return Energy mode used?
	 */
	virtual bool useEnergy() const;

	/*!
	 * Properties
	 */
	Property<int32> nbAdcControllers;		// Wanted number ADCs PHA
	Property<int32> nbBits; 				// Wanted resolution
	Property<int32> gateWidth; 				// Gate width
	Property<int32> scopeChannel; 			// Scope channel
	Property<int32> totalChannels; 			// Total nb of channels
	Property<int32> usePn1Coincidence;
	Property<int32> useWaveForm;
	Property<int32> useScope;
	Property<std::string> useScopeStr;

	Property<std::string> liveProcessType; 	// Live Process Type

	DynamicProperty<int32> boardIndex;			// RIO3 board index
	DynamicProperty<int32> channelsPerBoard;	// Total nb of channels per RIO3 board

	static const std::string NO_LIVE_PROCESS;
	static const std::string HISTOGRAM_PROCESS;
	static const std::string COINCIDENCE_PROCESS;

	std::vector<DriverPtr<acquisition::DPPAcquisitionCommon> > dppDriver;
	std::vector<ControllerPtr<CAENAdcController> > adcController;

private:
	virtual void refreshNDriversProperty(int32 aValue) throw (CannotSetValue);

	void updateCurve0ArrayProperty(int32 index);
	void updateCurve1ArrayProperty(int32 index);
	void updateCurve2ArrayProperty(int32 index);
	void updateCurve3ArrayProperty(int32 index);
	void updateCurve4ArrayProperty(int32 index);
	void updateAcqModeProperty(int32 index);
	void updateRecordLengthProperty(int32 index);

	void refreshNbAdcControllersProperty(int32 aValue) throw (CannotSetValue);
	virtual void refreshUseListModeProperty(bool aValue) throw (CannotSetValue);
	void refreshUsePn1CoincidenceModeProperty(int32 aValue) throw (CannotSetValue);
	void refreshNbBitsProperty(int32 value) throw (CannotSetValue);

	void refreshLiveProcessTypeProperty(const std::string& value) throw (CannotSetValue);

	/*!
	 * \brief Stop command
	 */
	virtual void stop();

	/*!
	 * \brief update property methods
	 */
//	void updateUseScopeProperty(DriverPtr<DPPAcquisitionCommon>& driver);
//	void updateNbCells(int32 index, DriverPtr<DPPAcquisitionCommon>& driver);

	void setScopeMode();
	void setChannelsPerBoard();
	void setCrateNumber();
	void setContext();

	ControllerPtr<DPPLiveProcess> m_histogramController;
	ControllerPtr<DPPLiveProcess> m_coincidenceController;
	ControllerPtr<DPPLiveProcess> m_liveProcessController;

	bool writeParamDone;
};

}
#endif
