/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "DielectricsRemoteWindeta.h"
#include "WindetaMessages.pb.h"
#include <boost/lexical_cast.hpp>
#include <common/base/ServerProperties.h>
#include <common/base/FileUtilities.h>
#include "controllers/common/family/Families.h"
#include "InstrumentManager/InstrumentManager.h"
#include "controllers/common/acquisition/Count.h"
#include "controllers/common/acquisition/ExperimentData.h"
#include <boost/filesystem/operations.hpp>
#include <thread>
#include <fstream>
#include <iostream>

namespace dielectrics {

using namespace std;
using namespace common;
using namespace boost;
using namespace cameo;

const std::string RemoteWindeta::TYPE = "dielectrics_remote_windeta";
const std::string RemoteWindeta::WINDETA_APPLICATION = "dielectrics-windeta";
const std::string RemoteWindeta::RESPONDER_NAME = "dielectrics_controller";
const std::string RemoteWindeta::PUBLISHER_NAME = "dielectrics_pub";
const std::string RemoteWindeta::INDEPENDENT_PREFIX = "windeta-";

const std::string RemoteWindeta::INDEPENDENT_VALUE = "independent";
const std::string RemoteWindeta::NUMOR_VALUE = "numor";
const std::string RemoteWindeta::NUMOR_1_VALUE = "numor+1";
const std::string RemoteWindeta::COUNT_SYNC_VALUE = "sync";

RemoteWindeta::RemoteWindeta(const std::string& name) :
	ExperimentController(name), controller::Stoppable(this), controller::Init(this), controller::Raz(this), Test(this),
	m_independentDataFileId(0), m_countNumor(0), m_stopped(false) {

	setFamily(family::ACQUISITION, family::POLARIZATIONS);

	serverEndpoint.init(this, SAVE, "cameo_server");
	initialized.init(this, NOSAVE, "initialized");
	applicationState.init(this, NOSAVE, "remote_state");

	preset.init(this, SAVE | SPY, "preset", "Preset");
	numberOfPoints.init(this, NOSAVE | SPY, "number_of_points", "Points");
	startTimeout.init(this, SAVE, "start_timeout");

	dataFilenameFormat.init(this, SAVE, "data_filename_format", "Filename");
	dataFilenameFormatSize.init(this, NOSAVE, "data_filename_format_size");
	dataFilenameFormatValues.init(this, NOSAVE, "data_filename_format_values");
	dataFilenameFormatLabels.init(this, NOSAVE, "data_filename_format_labels");
	dataFilenameFormatSize = 4;

	dataFilenameFormatValues.resize(dataFilenameFormatSize());
	dataFilenameFormatValues.set(0, NUMOR_VALUE);
	dataFilenameFormatValues.set(1, NUMOR_1_VALUE);
	dataFilenameFormatValues.set(2, INDEPENDENT_VALUE);
	dataFilenameFormatValues.set(3, COUNT_SYNC_VALUE);

	dataFilenameFormatLabels.resize(dataFilenameFormatSize());
	dataFilenameFormatLabels.set(0, "Numor");
	dataFilenameFormatLabels.set(1, "Numor+1");
	dataFilenameFormatLabels.set(2, "Independent");
	dataFilenameFormatLabels.set(3, "Sync");

	dataFilenameFormat.setEnumeratedValues(dataFilenameFormatValues);
	dataFilenameFormat.setEnumeratedLabels(dataFilenameFormatLabels);

	presetsChanged.init(this, NOSAVE, "presets_changed");
	presetsSize.init(this, NOSAVE, "presets_size");
	presets.init(this, NOSAVE, "presets");

	m_CountSpy.init(this, "acquisition_spy");
	m_driver.init(this, "driver");

	initIndependentDataFileId();
}

RemoteWindeta::RemoteWindeta(const RemoteWindeta& controller) :
	ExperimentController(controller), controller::Stoppable(this), controller::Init(this), controller::Raz(this), Test(this),
	m_independentDataFileId(0), m_countNumor(0), m_stopped(false) {
}

RemoteWindeta::~RemoteWindeta() {
	reset();
}

void RemoteWindeta::postConfiguration() {

	registerUpdater(m_CountSpy->statusMessage, &RemoteWindeta::updateStatusMessage, this);

	if (m_server.get() == nullptr) {
		// Create a new server if it is not already created.
		// In case of a simulated server, avoid the real remote endpoint.
		if (serverEndpoint() == "" || m_driver->state() != REAL_DEVICE_CONTAINER) {
			m_server.reset(new Server(application::This::getServer().getEndpoint()));
		}
		else {
			m_server.reset(new Server(serverEndpoint()));
		}
	}

	// Do not initialise if the controller is disabled.
	if (isEnabled()) {
		initialized = initApplication();
	}
}

void RemoteWindeta::setEnabled(bool all, bool postconf) {

	ExperimentController::setEnabled(all);
}

void RemoteWindeta::setDisabled(bool all) {

	ExperimentController::setDisabled(all);
}

bool RemoteWindeta::initApplication() {

	// Do not initialize if it is already done
	if (initialized()) {
		return true;
	}

	try {
		// Start the Windeta server
		if (!m_server->isAvailable(10000)) {
			cout << "Windeta server is not available" << endl;
			return false;
		}

		cout << "Windeta server is connected to " << getName() << endl;

		m_windetaApplication = m_server->connect(WINDETA_APPLICATION);

		cout << "Connected application" << endl;

		if (m_windetaApplication->exists()) {
			// The application exists from a previous server session
			m_windetaApplication->stop();
			application::State state = m_windetaApplication->waitFor();
			cout << "Terminated Windeta application " << state << endl;
		}

		m_windetaApplication = m_server->start(WINDETA_APPLICATION);

		if (!m_windetaApplication->exists()) {
			cout << "No Windeta application" << endl;
			return false;
		}

		applicationState = m_windetaApplication->waitFor(application::STARTING);

		// First delete the requester if it exists.
		m_requester.reset(nullptr);

		// Create a new requester.
		m_requester = application::Requester::create(*m_windetaApplication, RESPONDER_NAME);

		if (m_requester.get() == nullptr) {
			cout << "requester error" << endl;
			return false;
		}

		// Create a new subscriber.
		m_subscriber = application::Subscriber::create(*m_windetaApplication, PUBLISHER_NAME);

		// Start the subscriber loop
		m_subscriberThread.reset(new std::thread(std::bind(&RemoteWindeta::subscriberLoop, this)));

		// Application initialized.
		initialized = true;

		applicationState = m_windetaApplication->waitFor(application::RUNNING);

		cout << "Windeta application initialized" << endl;

		// Request for the presets.
		initPresets();

		return true;

	} catch (const std::exception & e) {
		// Currently an exception can occur during isAvailable.
		cout << "Unexpected exception during Windeta connection: " << e.what() << endl;
	}

	return false;
}

void RemoteWindeta::initPresets() {

	// Serialize the request.
	windeta::Request requestType;
	requestType.set_type(windeta::Request::PRESETS);

	// Send the message.
	m_requester->sendTwoBinaryParts(requestType.SerializeAsString(), "");

	// Wait for the response synchronously.
	// Note that responses can be processed asynchronously if the server is able to.
	string data;
	m_requester->receiveBinary(data);

	windeta::Response response;
	response.ParseFromString(data);
}

void RemoteWindeta::init() {

	// Do not initialise if the controller is disabled.
	if (isEnabled()) {
		initialized = initApplication();
	}
}

void RemoteWindeta::reset() {

	// Reset if the controller was initialized.
	if (initialized()) {

		// Stop the subscriber
		m_subscriber->cancel();

		// The subscriber thread terminates
		m_subscriberThread->join();

		// First stop the application.
		stopApplication();

		// Restart the application.
		initialized = false;
	}
}

void RemoteWindeta::updateStatusMessage() {

	if (m_CountSpy->statusMessage() == acquisition::Count::COUNTING_MESSAGE) {
		m_countNumor = m_CountSpy->numor();
	}
}

void RemoteWindeta::start() {

	if (!initApplication()) {
		return;
	}

	commandStatus.setIdle();

	// Check initial status.
	if (commandStatus.isRunning()) {
		log(Level::s_Error) << name << "cannot start because it is already running" << endlog;
		return;
	}

	log(Level::s_Debug) << name << cursor << "preset " << preset() << endlog;

	int timeout = startTimeout();
	int i = 0;

	m_stopped = false;
	m_countNumor = 0;

	// Wait for the count to have changed the numor so that we can use it in case of sync.
	if (dataFilenameFormat() == COUNT_SYNC_VALUE) {

		// Poll in milliseconds.
		while (m_countNumor == 0 && i < timeout * 100 && !m_stopped) {
			std::this_thread::sleep_for(std::chrono::milliseconds(10));
			i++;
		}

		if (m_stopped) {
			log(Level::s_Info) << name << "preset " << preset() << " stopped" << endlog;
			return;
		}

		if (m_countNumor == 0) {
			log(Level::s_Error) << name << "preset " << preset() << " start timeout" << endlog;
			return;
		}
	}

	// Serialize the request.
	windeta::Request requestType;
	requestType.set_type(windeta::Request::START_MEASURE);

	windeta::StartMeasure requestMessage;
	requestMessage.set_preset(preset());

	// Send the message.
	m_requester->sendTwoBinaryParts(requestType.SerializeAsString(), requestMessage.SerializeAsString());

	// Wait for the response synchronously.
	// Note that responses can be processed asynchronously if the server is able to.
	string data;
	m_requester->receiveBinary(data);

	windeta::Response response;
	response.ParseFromString(data);

	// Set the data filename.
	setDataFilename();

	// The status must become running within a timeout.
	i = 0;
	while (!commandStatus.isRunning() && !commandStatus.isError() && i < timeout && !m_stopped) {
		std::this_thread::sleep_for(std::chrono::seconds(1));
		i++;
	}

	if (!commandStatus.isRunning() && !commandStatus.isError() && !m_stopped) {
		log(Level::s_Error) << name << "preset " << preset() << " to " << m_dataFilename << " start timeout" << endlog;
		return;
	}

	// Loop on the status.
	while (commandStatus.isRunning() && !m_stopped) {
		std::this_thread::sleep_for(std::chrono::seconds(1));
	}

	// Check the execution.
	if (m_stopped) {
		log(Level::s_Info) << name << "preset " << preset() << " to " << m_dataFilename << " stopped" << endlog;
	}
	else if (commandStatus.isError()) {
		log(Level::s_Error) << name << "preset " << preset() << " to " << m_dataFilename << " on error" << endlog;
	}
	else {
		log(Level::s_Info) << name << "preset " << preset() << " to " << m_dataFilename << " finished" << endlog;
	}

}

void RemoteWindeta::stop() {

	if (!initApplication()) {
		return;
	}

	m_stopped = true;
}

void RemoteWindeta::raz() {

	stopApplication();
}

void RemoteWindeta::test() {

	if (!initApplication()) {
		return;
	}

	// Serialize the request.
	windeta::Request requestType;
	requestType.set_type(windeta::Request::TEST);

	// Send the message.
	m_requester->sendTwoBinaryParts(requestType.SerializeAsString(), "");

	cout << getName() << " sent request " << static_cast<int>(requestType.type()) << endl;

	// Wait for the response synchronously.
	// Note that responses can be processed asynchronously if the server is able to.
	string data;
	m_requester->receiveBinary(data);

	windeta::Response response;
	response.ParseFromString(data);

	cout << "received response " << response.errorcode() << endl;
}

void RemoteWindeta::stopApplication() {

	if (!initialized()) {
		return;
	}

	// Exit immediately if the server is not available.
	if (!m_server->isAvailable(10000)) {
		return;
	}

	// Serialize the request.
	windeta::Request requestType;
	requestType.set_type(windeta::Request::STOP_SERVER);

	// Send the message.
	m_requester->sendTwoBinaryParts(requestType.SerializeAsString(), "");

	// Wait for the response synchronously.
	// Note that responses can be processed asynchronously if the server is able to.
	string data;
	m_requester->receiveBinary(data);

	windeta::Response response;
	response.ParseFromString(data);

	// Stop the remote application.
	m_windetaApplication->stop();

	// Wait for the termination
	applicationState = m_windetaApplication->waitFor();

	cout << "Windeta application terminated with state " << application::toString(applicationState()) << endl;

	// Reset initialized.
	initialized = false;
}

void RemoteWindeta::initIndependentDataFileId() {

	string dataPath = ServerProperties::getInstance()->getNomadDataPath();

	vector<filesystem::path> files = FileUtilities::getFilePathsWithExtension(dataPath, "diel");

	for (auto f : files) {

		string filename = f.filename().string();
		if (filename.find(INDEPENDENT_PREFIX) != string::npos) {

			int length = filename.length() - 5 - INDEPENDENT_PREFIX.length();
			string idString = filename.substr(INDEPENDENT_PREFIX.length(), length);

			int id = lexical_cast<int>(idString);

			if (id > m_independentDataFileId) {
				m_independentDataFileId = id;
			}
		}
	}

	cout << "Independent data file id is " << m_independentDataFileId << endl;
}

void RemoteWindeta::setDataFilename() {

	// Find the experiment data controller.
	acquisition::ExperimentData* experimentData = nullptr;

	try {
		string controllerName = InstrumentManager::getInstance()->getControllersWithType("title").at(0);
		experimentData = dynamic_cast<acquisition::ExperimentData*>(InstrumentManager::getInstance()->getExperimentController(controllerName));
	}
	catch (...) {
		log(Level::s_Error) << name << "cannot find a name for data file associated to a numor" << endlog;
	}

	// Set the filename according to the format.
	if (dataFilenameFormat() == INDEPENDENT_VALUE || experimentData == nullptr) {

		m_independentDataFileId++;
		m_dataFilename = ServerProperties::getInstance()->getNomadDataPath() + INDEPENDENT_PREFIX + lexical_cast<string>(m_independentDataFileId) + ".diel";
	}
	else {
		// Set the filename related to numor.
		int numor = 0;
		if (dataFilenameFormat() == NUMOR_VALUE || dataFilenameFormat() == NUMOR_1_VALUE) {
			numor = experimentData->numor();
		}
		else {
			numor = m_countNumor;
		}

		if (dataFilenameFormat() == NUMOR_1_VALUE) {
			numor++;
		}

		stringstream numorValue;
		numorValue << boost::format("%06d") % numor;

		string dataPath = ServerProperties::getInstance()->getNomadDataPath();
		string baseDataFilename = dataPath + numorValue.str();
		string dataFilename = baseDataFilename + ".diel";

		filesystem::path file(dataFilename);
		if (filesystem::exists(file)) {
			int i = 1;
			while (true) {
				dataFilename = baseDataFilename + "-" + lexical_cast<string>(i) + ".diel";
				filesystem::path file(dataFilename);

				if (!filesystem::exists(file)) {
					break;
				}

				i++;
			}
		}

		m_dataFilename = dataFilename;
	}
}

void RemoteWindeta::saveData(const std::string& data) {

	ofstream dataFile(m_dataFilename.c_str());
	if (dataFile.is_open()) {
		dataFile << data;
		dataFile.close();
	}
	else {
		log(Level::s_Error) << name << "preset " << preset() << " to " << m_dataFilename << " fails to write data" << endlog;
	}
}

void RemoteWindeta::subscriberLoop() {

	// Loop on events
	string data1;
	string data2;

	while (m_subscriber->receiveTwoBinaryParts(data1, data2)) {

		windeta::Event event;
		event.ParseFromString(data1);

		if (event.type() == windeta::Event::STATUS) {
			windeta::Status status;
			status.ParseFromString(data2);

			if (status.type() == windeta::Status::STARTED) {
				log(Level::s_Debug) << name << " started" << endlog;

				commandStatus.setRunning();
			}
			else if (status.type() == windeta::Status::FINISHED) {
				log(Level::s_Debug) << name << " finished" << endlog;

				commandStatus.setIdle();
			}
			else if (status.type() == windeta::Status::ERROR) {
				log(Level::s_Debug) << name << " error" << endlog;

				commandStatus.setError();
			}
		}
		else if (event.type() == windeta::Event::DATA) {
			windeta::Data data;
			data.ParseFromString(data2);

			numberOfPoints = data.points();
			saveData(data.data());
		}
		else if (event.type() == windeta::Event::PRESETS) {
			windeta::Presets messagePresets;
			messagePresets.ParseFromString(data2);

			cout << "Presets" << endl;

			int size = messagePresets.list_size();

			presets.resize(size);

			for (int i = 0; i < size; ++i) {
				presets.set(i, messagePresets.list(i));

				cout << "  " << presets.get(i) << endl;
			}

			presetsSize = size;
			presetsChanged = presetsChanged() + 1;
		}
	}
}

}
