/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "IKonDriver.h"

#include <math.h>

#include "IKonDef.h"
#include "IKonState.h"
#include "PerfectIKonDriver.h"
#include "RealIKonDriver.h"

using namespace std;

namespace ikon {

const std::string IKonDriver::TYPE = "IKon";
const std::string IKonDriver::ON_COMMAND = "on";
const std::string IKonDriver::OFF_COMMAND = "off";

/*
 * Constructor
 */
IKonDriver::IKonDriver(const string& name) :
				acquisition::NoTofAcquisitionCommon(name),acquisition::KineticAcquisitionCommon(name),
				acquisition::TimeGateCommon(name),
				acquisition::DetectorCommon(name),temperature::TemperatureCommon(name) {

	registerStates(new RealIKonDriver(this), new PerfectIKonDriver(this), new PerfectIKonDriver(this));

	/*
	 * Init the device command list
	 */
	height.init(this, SAVE, "height");
	width.init(this, SAVE, "width");
	hbind.init(this, SAVE, "hbind");
	vbind.init(this, SAVE, "vbind");
	hflip.init(this, SAVE, "hflip");
	vflip.init(this, SAVE, "vflip");
	rotate.init(this, SAVE, "rotate");
	acqmode.init(this, SAVE, "acqmode");
	cooling.init(this, SAVE, "cooling");
	readout.init(this, SAVE, "readout");
	shutopentime.init(this, SAVE, "shutopentime");
	shutclosetime.init(this, SAVE, "shutclosetime");
	shutmode.init(this, SAVE, "shutmode");
	hsspeed.init(this, SAVE, "hhspeed");
	gain.init(this, SAVE, "gain");
	bclamp.init(this, SAVE, "bclamp");
	tempstatus.init(this, SAVE, "tempstatus");
	modeFVB.init(this, SAVE, "modeFVB");;
	vsspeed.init(this, SAVE, "vsspeed");

	initCommand(driver::INIT_COMMAND);
	initCommand(driver::READ_COMMAND);
	initCommand(driver::SYNCHRONIZE_READ_COMMAND);
	initCommand(driver::START_COMMAND);
	initCommand(driver::RESUME_COMMAND);
	initCommand(driver::PAUSE_COMMAND);
	initCommand(driver::STOP_COMMAND);
	initCommand(driver::STATUS_COMMAND);
	initCommand(driver::READ_INFOS_COMMAND);
	initCommand(DetectorCommon::CLEAR_COMMAND);
	initCommand(DetectorCommon::REGROUP_COMMAND);
	initCommand(AcquisitionCommon::WRITE_PARAMETERS_COMMAND);
	initCommand(driver::INIT_COMMAND);
	initCommand(ON_COMMAND);
	initCommand(OFF_COMMAND);

	// Init functions
	registerParentFunction(NONE_FUNCTION);
	registerFunction(NONE_FUNCTION);

	deviceType = ROOT_DEVICE_TYPE_DEVICE_CONTAINER;
	m_SynchroniseReadActivated = false;
	m_SynchroniseReadTerminated = false;
	m_DataShort = NULL;

	registerObserverCommand(COMMAND_STATUS_DEVICE_CONTAINER, 10);
	registerObserverCommand(COMMAND_READ_DEVICE_CONTAINER, 100);
	registerSpyCommand(COMMAND_READ_DEVICE_CONTAINER, 5);
}

/*
 * Destructor
 */
IKonDriver::~IKonDriver() {
}

/*
 * refreshDataSizeProperty
 */
void IKonDriver::refreshDataSizeProperty(int32 value) throw (CannotSetValue) {
	if (m_DataShort != NULL) {
		delete[] m_DataShort;
	}
	m_DataShort = new uint16[value];
	if (m_DataShort == NULL) {
		sendErrorEvent(NOTENOUGH_MEMORY_ERROR, "Can't create data table");
		throw CannotSetValue();
	}
	DetectorCommon::refreshDataSizeProperty(value);
}

/*
 * refreshModeProperty
 */
void IKonDriver::refreshModeProperty(int32 value) throw (CannotSetValue) {
	if (value != acquisition::GateCommon::TIME_MODE) {
		throw CannotSetValue();
	}
}

/*
 * execute
 */
void IKonDriver::execute(const std::string& aCommand) {

	if ((aCommand != driver::PAUSE_COMMAND) && (aCommand != driver::STOP_COMMAND) && (aCommand != driver::STATUS_COMMAND)
			&& (aCommand != driver::READ_COMMAND)) {
		commandProgression = PROGRESSION_UNKNOWNSTATE_DEVICE_CONTAINER;
	}
	IKonState* currentState = dynamic_cast<IKonState*>(getCurrentState());
	// Check command
	if (aCommand == driver::INIT_COMMAND) {
		// Init command
		currentState->init();
	} else if (aCommand == driver::READ_COMMAND) {
		// Read command
		currentState->read();
	} else if (aCommand == driver::SYNCHRONIZE_READ_COMMAND) {
		// Read command
		m_SynchroniseReadActivated = true;
		m_SynchroniseReadTerminated = false;
		currentState->synchroniseRead();
	} else if (aCommand == driver::START_COMMAND) {
		// Start command
		startActivated = true;
		currentState->start();
	} else if (aCommand == driver::RESUME_COMMAND) {
		// Continue command
		currentState->resume();
	} else if (aCommand == driver::PAUSE_COMMAND) {
		// Stop command
		currentState->pause();
	} else if (aCommand == driver::STOP_COMMAND) {
		// Stop command
		currentState->stop();
		startActivated = false;
	} else if (aCommand == driver::STATUS_COMMAND) {
		// Status command
		currentState->readStatus();
	} else if (aCommand == driver::READ_INFOS_COMMAND) {
		// Info command
		currentState->readInfos();
	} else if (aCommand == DetectorCommon::CLEAR_COMMAND) {
		// Move command
		sum = 0;
		currentState->clear();
	} else if (aCommand == AcquisitionCommon::WRITE_PARAMETERS_COMMAND) {
		// Write command
		currentState->writeParam();
	} else if (aCommand == ON_COMMAND) {
		// on command
		currentState->setCooler(1);
	} else if (aCommand ==OFF_COMMAND) {
		// off command
		currentState->setCooler(0);
	} else {
		//Error bad command
		//Todo , normaly it will be detect it device action driver
	}
}

/*
 * calculateProgression
 */
void IKonDriver::calculateProgression() {
	int32 progression = 0;
	float64 setpoint = 0.0;
	float64 actual = 0.0;

	setpoint = time.setpoint();
	actual = time();

	if (setpoint == 0.0) {
		commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
	} else {
		progression = (int32) (actual / setpoint * PROGRESSION_RATIO_DEVICE_CONTAINER);
		// Be sure that progression > 0%
		if (progression < PROGRESSION_UNKNOWNSTATE_DEVICE_CONTAINER)
			progression = 0;
		// Be sure that only Phased status put 100%
		if (progression >= PROGRESSION_END_DEVICE_CONTAINER)
			progression = PROGRESSION_END_DEVICE_CONTAINER - 1;
		commandProgression = progression;
	}
}

/*
 * IKonDriver
 */
void IKonDriver::computeGateStatus(int32 value) {
	int32 valueout = 0;
	if ((value == ikon::EXPOSURE_IN_PROGRESS) || (value == ikon::READOUT_IN_PROGRESS)
			|| (value == ikon::ACQUISITION_IN_PROGRESS)) {
		valueout |= GateCommon::RUNNING_STATUS;
	} else if (value == ikon::READOUT_COMPLETE) {
		valueout |= GateCommon::END_STATUS;
	}
	gateStatus.update(valueout);
}

/*
 * IKonDriver
 */
void IKonDriver::computeDetectorStatus(int32 value) {
	int32 valueout = 0;
	detectorStatus.update(valueout);

}

/*
 * IKonDriver
 */
void IKonDriver::computeAcquisitionStatus(int32 value) {
	int32 valueout = 0;
	acquisitionStatus.update(valueout);
}

}

