/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include "RealCAENQdcDriver.h"

#include <controllers/common/acquisition/mode/DPPAcquisitionCommon.h>
#include "CAENQdcDriver.h"

using namespace std;

namespace caen_qdc {

/*
 * Constructor
 */
RealCAENQdcDriver::RealCAENQdcDriver(CAENQdcDriver* owner) : CAENQdcState(owner) {

	m_boardHandle = 0;
	ret = CAEN_DGTZ_Success;
}

/*
 * Destructor
 */
RealCAENQdcDriver::~RealCAENQdcDriver() {

}

/*!
 * \brief open command implementation
 */
void RealCAENQdcDriver::open() {

	boardParams.LinkType = (CAEN_DGTZ_ConnectionType)owner()->linkType();
	boardParams.VMEBaseAddress = owner()->address();
	ret = CAEN_DGTZ_OpenDigitizer((CAEN_DGTZ_ConnectionType)boardParams.LinkType, 0, 0, boardParams.VMEBaseAddress, &m_boardHandle);
	if (ret) {
		cerr << "Error CAEN_DGTZ_OpenDigitizer: " << ret << endl;
		return;
	}
}

/*!
 * \brief close command implementation
 */
void RealCAENQdcDriver::close() {

	CAEN_DGTZ_CloseDigitizer(m_boardHandle);
}

/*!
 * \brief Init command implementation
 */
void RealCAENQdcDriver::init() {

    reset();

	readInfos();

	open();

    ret = CAEN_DGTZ_Success;

    uint32_t GroupMask = 0xFF;

    // Activate all groups
    for(int group=0; group<8; group++) {
       	ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x10A8 + 0x100*group, 0xFF);
    }

	// Set I/O level
	ret |= CAEN_DGTZ_SetIOLevel(m_boardHandle, (CAEN_DGTZ_IOLevel_t)owner()->ioLevel());

	/* Disable Group self trigger for the acquisition (mask = 0) */
	ret |= CAEN_DGTZ_SetGroupSelfTrigger(m_boardHandle, CAEN_DGTZ_TRGMODE_ACQ_ONLY,0x00);

	/* Set the behaviour when a SW trigger arrives */
	ret |= CAEN_DGTZ_SetSWTriggerMode(m_boardHandle, CAEN_DGTZ_TRGMODE_ACQ_ONLY);

	/* Set the max number of events/aggregates to transfer in a sigle readout */
	ret |= CAEN_DGTZ_SetMaxNumAggregatesBLT(m_boardHandle, MAX_AGGR_NUM_PER_BLOCK_TRANSFER);

	/* Set the start/stop acquisition control and arm acquisition */
    ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8100, 0x5);

	if (owner()->triggerMode() == 2) {
		uint32_t d32;
		owner()->selfTrigger = 1;
		ret |= CAEN_DGTZ_ReadRegister(m_boardHandle, 0x8000, &d32);
		d32 = d32 | (0 & 0x3 << 20);
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8000, d32);
	} else {
		owner()->selfTrigger = 0;
	}

    /* Set Baseline (used in fixed baseline mode only) */
	ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8038, owner()->fixedBaseline());

    /* enable Charge mode */
    ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8004, 0x00080000);

    /* enable Timestamp */
    ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8004, 0x00040000);

    /* Set Extended Time Stamp if enabled*/
    ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8004, 1 << 17);

    /* Check errors */
	if(ret != CAEN_DGTZ_Success) {
        cout << "Errors during Digitizer Initialization" << endl;
        return;
    }

	close();

	writeParam();
}

int32_t RealCAENQdcDriver::setChannelGroupMask(uint32_t group, uint32_t channelMask) {

	int ret = CAEN_DGTZ_Success;

    ret = setGroupTriggerMask(group, channelMask);

    return ret;
}

int32_t RealCAENQdcDriver::setGroupTriggerMask(uint32_t group, uint32_t channelMask){

    int32_t address;

    if (group > V1740_MAX_CHANNELS -1) return CAEN_DGTZ_InvalidChannelNumber;
    address = getChannelAddress(CAEN_DGTZ_CHANNEL_GROUP_V1740_BASE_ADDRESS, (uint8_t) group);
    ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, address, channelMask);

    return ret;
}

/*!
 * \brief Clear command implementation
 */
void RealCAENQdcDriver::clear() {

	open();
	ret = CAEN_DGTZ_ClearData(m_boardHandle);
	if (ret) {
		cerr << "Error CAEN_DGTZ_ClearData: " << ret << endl;
	}
	close();
}

/*!
 * \brief Reset command implementation
 */
void RealCAENQdcDriver::reset() {

	open();
	ret = CAEN_DGTZ_Reset(m_boardHandle);
	if (ret) {
		cerr << "Error CAEN_DGTZ_Reset: " << ret << endl;
	}
	close();
}

/*!
 * \brief Write Param command implementation
 */
void RealCAENQdcDriver::writeParam() {

    uint32_t DppCtrl1 = 0;

    ret = CAEN_DGTZ_Success;

    open();

    DppCtrl1 = (((owner()->triggerHysteresis() & 0x1)   << 30)   |
                ((owner()->selfTrigger() & 1)     		<< 24)   |
                ((owner()->baselineMean() & 0x7)     	<< 20)   |
                ((owner()->triggerMode() & 3)           << 18)   |
                ((owner()->chargeSensitivity() & 0x7)	<<  0)   |
                ((owner()->pulsePolarity() & 1)         << 16)   |
                ((owner()->chargePedestal() & 1)        <<  8)   |
                ((owner()->testPulseRate() & 3)     	<<  5)   |
                ((owner()->testPulseEnable() & 1)       <<  4)   |
                ((owner()->inputSmoothing() & 7)       	<< 12));

    ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8040, DppCtrl1);

    /* Set the waveform lenght (in samples) */
    ret |= setRecordLength(owner()->recordLength());

    /* set Scope mode */
    if (owner()->acqMode() == 0)
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8004, 0x00010000);
	else
        ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8008, 0x00010000);

    /* Set number of events per memory buffer */
    //ret |= setNumEvAggregate(MAX_AGGR_NUM_PER_BLOCK_TRANSFER);
    ret |= setNumEvAggregate(0);

    for(int group=0; group<8; group++) {
        /* Trigger Hold Off */
    	ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1074 + 0x100*group, owner()->triggerHoldOff.get(group));

    	/* Set Pre Trigger (in samples) */
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x103C + 0x100*group, owner()->preTrigger.get(group));

		/* Set Gate Offset (in samples) */
		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1034 + 0x100*group, owner()->preGate.get(group));

		/* Set Gate Width (in samples) */
    	 ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1030 + 0x100*group, owner()->gateWidth.get(group));

    	/* Set DC offset */
 		ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x1098 + 0x100*group, owner()->dcOffset.get(group));
    }

    // Set selfTrigger threshold
	for(int channel=0; channel<64; channel++)
		ret |= setChannelTriggerThreshold(channel, owner()->triggerThreshold.get(channel));

    /* Check errors */
	if(ret != CAEN_DGTZ_Success) {
		cout << "Errors during writeParam" << endl;
	}

	close();
}

/*!
 * \brief Read Infos command implementation
 */
void RealCAENQdcDriver::readInfos() {

	open();

	ret = CAEN_DGTZ_GetInfo(m_boardHandle, &boardInfo);
	if (ret) {
		cerr << "Error CAEN_DGTZ_GetInfo: " << ret << endl;
	}

	close();

	owner()->nbBits = boardInfo.ADC_NBits;
	owner()->model = boardInfo.Model;
	owner()->channels = boardInfo.Channels * 8; // nb of group channels with 8 channels per group
	owner()->serialNumber = boardInfo.SerialNumber;
	owner()->modelName = boardInfo.ModelName;
	owner()->rocFirmware = boardInfo.ROC_FirmwareRel;
	owner()->amcFirmware = boardInfo.AMC_FirmwareRel;
}

int32_t RealCAENQdcDriver::getChannelAddress(uint32_t base, uint16_t channel) {

	//return base + (0x100 * channel);
    int chShift = 8;
    uint32_t res = base & ~(0xF<<chShift); // set bits [11:8] to zero
    res |= channel << chShift;

    return res;
}

int32_t RealCAENQdcDriver::setChannelTriggerThreshold(uint32_t channel, uint32_t Tvalue) {

	int gr;
    int ch;

    uint32_t addr;

    if (channel > 64)
        return -1;

    gr = channel/8;
    ch = channel % 8;

    addr = 0x1000 + 0x100*gr + 0xD0 + 4*ch;
    ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, addr, Tvalue);

    return ret;
}

int32_t RealCAENQdcDriver::setRecordLength(uint32_t recordLength) {

	ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8024, recordLength/8);

    return ret;
}

int32_t RealCAENQdcDriver::setNumEvAggregate(uint32_t nEvAggr) {

	uint32_t d32;

    if (nEvAggr == 0) {
        CAEN_DGTZ_ReadRegister(m_boardHandle, 0x8000, &d32);
        if (d32 & (1<<16)) {                                          		 /* waveform enabled      */
        	ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x800C, 0x1);      /* Buffer organization   *///0x3
            ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8020, 1);        /* Events per aggregate  */
        } else {
            ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x800C, 0xA);      /* Buffer organization   */
            ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8020, 1);       /* Events per aggregate  */
        }
    } else {
        ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x800C, 0x1);          /* Buffer organization   */
        ret |= CAEN_DGTZ_WriteRegister(m_boardHandle, 0x8020, nEvAggr);      /* Events per aggregate  */
    }

    return ret;
}

void RealCAENQdcDriver::setAcqMode() {

}

}
