/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <boost/thread/thread.hpp>
#include <drivers/ioxos/cadaracheacq/CadaracheAcqDef.h>
#include <drivers/ioxos/cadaracheacq/SimulatedCadaracheAcqDriver.h>

using namespace acquisition;
using namespace std;

namespace cadaracheacq {

/*
 * Constructor
 */
SimulatedCadaracheAcqDriver::SimulatedCadaracheAcqDriver(CadaracheAcqDriver* owner) :
		CadaracheAcqState(owner)

{
	m_Stop = false;
	m_Pause = false;
	m_Finished = true;
	m_Status = 0;
}

/*
 * Destructor
 */
SimulatedCadaracheAcqDriver::~SimulatedCadaracheAcqDriver() {
	/* Empty */
}

/*
 * init
 */
void SimulatedCadaracheAcqDriver::init() {
}

/*
 * clear
 */
void SimulatedCadaracheAcqDriver::clear() {
}

/*
 * writeParam
 */
void SimulatedCadaracheAcqDriver::writeParam() {
}

/*
 * readInfos
 */
void SimulatedCadaracheAcqDriver::readInfos() {
	uint32 version = 0xDC010811;
	ostringstream buf;
	buf << hex << version << dec;
	owner()->version.update(buf.str());
}

/*
 *stop
 */
void SimulatedCadaracheAcqDriver::start() {
	// Start command
	m_Times.clear();
	m_Finished = false;
	m_Stop = false;
	m_Pause = false;

	float64* times = owner()->slicesAcq();
	int32 nb = owner()->nbSlicesAcq();
	float64 wanted_time = 0;
	for (int32 i = 0; i < nb; i++) {
		wanted_time += times[i] / 1.0e6;
		m_Times.push_back(times[i] / 1.0e6);
	}
	owner()->time.setpoint.update(wanted_time * (float64) owner()->nbRepaint.setpoint());

	m_Status = RUNNING_STATUS;
	// Launch thread timer
	boost::thread td(boost::bind(&SimulatedCadaracheAcqDriver::timer, this));
}

/*
 *stop
 */
void SimulatedCadaracheAcqDriver::pause() {
	m_Pause = true;
	m_TimeCounter.pause();
//	m_Status = COUNT_PAUSED;
}

/*
 *resum
 */
void SimulatedCadaracheAcqDriver::resume() {
	m_TimeCounter.resume();
	m_Pause = false;
	m_Status = RUNNING_STATUS;
}

/*
 *stop
 */
void SimulatedCadaracheAcqDriver::stop() {
	m_Stop = true;
	m_Pause = false;
	m_Status = 0;
}

/*
 * readParam
 */
void SimulatedCadaracheAcqDriver::readParam() {
}

/*
 * synchroniseRead
 */
void SimulatedCadaracheAcqDriver::synchroniseRead() {
	read();
	owner()->m_SynchroniseReadTerminated = true;
}

/*
 * read
 */
void SimulatedCadaracheAcqDriver::read() {
	float64 value = m_TimeCounter.getTime();
	owner()->time.update(value);
	float64 sum = 0;
	int32* data = owner()->data();
	for(int32 i = 0;i < owner()->nbSlicesAcq();++i) {
		data[i] += 1;
		sum += data[i];
	}
	owner()->sum.update(sum);
	owner()->data.sendEvent();
}

/*
 * timer
 */
void SimulatedCadaracheAcqDriver::timer() {
	float64 tottime = 0;
	m_TimeCounter.clear();
	float64 time = 0;
	for (int32 j = 0; j < owner()->nbRepaint.setpoint(); ++j) {
		owner()->nbRepaint = j + 1;
		for (int32 i = 0; i < m_Times.size(); ++i) {
			owner()->actualSlice = i + 1;
			time = m_Times[i];
			tottime += time;
			while (m_TimeCounter.getTime() <= tottime) {
				usleep(10);
				if (m_Stop == true) {
					break;
				}
			}
			if (m_Stop == true) {
				break;
			}
		}
		if (m_Stop == true) {
			break;
		}
	}
	m_Finished = true;
}

/*
 * readStatus
 */
void SimulatedCadaracheAcqDriver::readStatus() {
	if (owner()->startActivated() == true) {
		if (m_Pause == false) {
			if (m_Finished == true) {
				m_Status = 0;
				owner()->status.update(m_Status);
				owner()->computeGateStatus(m_Status);
				owner()->startActivated = false;
				owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
			} else {
				owner()->status.update(m_Status);
				owner()->computeGateStatus(m_Status);
				owner()->calculateProgression();
			}
		}
	} else {
		if (owner()->m_SynchroniseReadActivated == true) {
			if (owner()->m_SynchroniseReadTerminated == true) {
				owner()->m_SynchroniseReadActivated = false;
				owner()->m_SynchroniseReadTerminated = false;
				owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
			}
		} else {
			owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
		}
		m_Status = 0;
		owner()->status.update(m_Status);
		owner()->computeGateStatus(m_Status);
	}
}

}
